"Test for samples module"
import numpy as np
from classifier.samples import Samples, TimeSeries


def test_impute_timeseries_AsExpected(
        setup_samples_df_with_nans_ref, setup_samples_df_with_nans):
    ref_samples = setup_samples_df_with_nans_ref
    timeseries = TimeSeries(None, None, None)

    timeseries_imputed = timeseries.impute_ts(setup_samples_df_with_nans)
    assert np.array_equal(
        ref_samples.to_numpy(),
        timeseries_imputed.to_numpy())


def test_outlier_removal_RemovedAsExpected(
        setup_samples_df, setup_samples_df_with_outliers):
    ref_samples = setup_samples_df
    samples_outlier_removed = Samples.outlier_removal(
        setup_samples_df_with_outliers)
    np.array_equal(ref_samples.values, samples_outlier_removed.values)



"Prediction Module"
import logging
import os
#from multiprocessing import process
from pathlib import Path
from typing import List, Tuple

import numpy as np
import rasterio
from rasterio.windows import Window

from classifier.samples import Samples, TimeSeries
from classifier.settings import US_ALGORITHMS
from classifier.utils.config import Configuration
from classifier.utils.general import write_tifs

# need to set the temp dir for multiprocessing in a container
#process.current_process()._config['tempdir'] = '/tmp/'

# Load all rasters in a windowed manner (in a vrt) and do the prediction
PREDICT_LOGGER = logging.getLogger(__name__)


def prediction_class_probabilities(
        predicted: np.ndarray, valid: np.ndarray, meta: dict,
        config: Configuration, window: Window) -> None:
    """Gets probability for the classified class and writes the tif for a
    windows

    Args:
        predicted (np.ndarray): The probabilities for all classes
        valid (np.ndarray): Array with valid values for prediction
        meta (dict): rasterio meta for writing tif
        config (Configuration)class): Contains config
        window (rasterio.Window) to predict and write

    """
    labels_internal_proba = (np.nanmax(predicted, axis=1)
                             .astype(np.float32))
    result_proba = np.full(valid.shape, -9999, np.float32)
    result_proba[valid] = labels_internal_proba
    write_tifs(os.path.join(config.tmp_dir.name, 'probability'),
               window,
               meta['dst_proba_meta'],
               result_proba)


def prediction_all_probabilities(
        predicted: np.ndarray,
        valid: np.ndarray,
        labels: dict,
        meta: dict,
        config: Configuration,
        window: Window) -> None:
    """Gets probability for the classified class and writes the tif for a
    windows

    Args:
        predicted (np.ndarray): The probabilities for all classes
        valid (np.ndarray): Array with valid values for prediction
        labels (dict): Conversion dictionary for labels
        meta (dict): rasterio meta for writing tif
        config (Configuration)class): Contains config
        window (rasterio.Window): processing window

    """
    # loop over the classes and save them all separately
    for internal_label in range(predicted.shape[1]):
        probabilities = predicted[:, internal_label]
        result_proba = np.full(valid.shape, -9999, np.float32)
        result_proba[valid] = probabilities
        write_tifs(
            os.path.join(config.tmp_dir.name, str(labels[internal_label])),
            window,
            meta['dst_proba_meta'],
            result_proba)


def do_prediction_proba_su(
        data: np.ndarray, valid: np.ndarray, model_dict: dict,
        config: Configuration, meta: dict, window: Window) -> np.ndarray:
    """Does the actual prediction and probability description for supervised
    methods

    Args:
        data (np.ndarray): Features
        valid (np.ndarray): Valid values for prediction
        model_dict (dict): The model dictionary containing model and
            algorithm info
        config (Configuration): The cli configuration
        meta (dict): rasterio meta for writing tif
        window (rasterio.window): asterio Window for which to do the prediction

    Returns:
        The predictions array

    """
    predicted = model_dict['model'].predict_proba(data)
    # Get the final indices of the highest probabilities
    labels_external = np.nanargmax(predicted, axis=1)

    # Label keys to int
    model_dict['labels'] = {
        int(k): int(v) for k, v in
        model_dict['labels'].items()
    }
    # Indices need to be transformed to the proper labels
    labels_transformed = np.vectorize(
        model_dict['labels'].get)(labels_external)
    result = np.full(valid.shape, -9999, np.int32)
    result[valid] = labels_transformed

    # See if we need to output some of the modelled probabilities
    if config.supervised.probability:
        prediction_class_probabilities(
            predicted,
            valid,
            meta,
            config,
            window
        )
    if config.supervised.all_probabilities:
        prediction_all_probabilities(
            predicted,
            valid,
            model_dict['labels'],
            meta,
            config,
            window
        )

    return result


def do_prediction_proba(
        data: np.ndarray, valid: np.ndarray, model_dict: dict,
        config: Configuration, meta: dict, window: Window) -> np.ndarray:
    """Calls the actual probability functions for (un)supervised
    methods

    Args:
        data (np.ndarray): Features
        valid (np.ndarray): Valid values for prediction
        model_dict (dict): The model dictionary containing model and
        algorithm info
        config (Configuration): The cli configuration
        meta (dict): rasterio meta for writing tif
        window: The rasterio Window for which to do the prediction

    Returns:
        The probability array

    """
    if config.app.algorithm == 'singleclass':
        result = predict_single_class(
            data,
            valid,
            model_dict,
            config,
            meta,
            window
        )
        return result
    return do_prediction_proba_su(data,
                                  valid,
                                  model_dict,
                                  config,
                                  meta,
                                  window)


def gather_data_for_prediction(
        rasters: List[Path],
        meta: dict,
        config: Configuration,
        window: Window) -> Tuple[np.ndarray, np.ndarray]:
    """Gathers all the data neccessary for a prediction using a raster TS

        Args:
            raster (List[Path]): Rasters
            meta (dict): rasterio meta for writing tif
            prediction_dict (dict): Parameters for the prediction
            window (rasterio Window): window of the chunk to predict
        Returns:
            data (np.array): Gathered and imputed data
            valid (np.array): Array with indices of locations of valid data
     """
    # Get all data in a df with pixels as columns and dates and bands as the
    # indices
    with rasterio.open(rasters[0], 'r') as src:
        shape = src.read(window=window).shape
        x_min, y_min, x_max, y_max = src.window_bounds(window)
        for meta_type in ['dst_meta', 'dst_proba_meta']:
            meta[meta_type].update(
                width=shape[2],
                height=shape[1],
                transform=src.window_transform(window))

    window_polygon = [[(x_min, y_min),
                       (x_min, y_max),
                       (x_max, y_max),
                       (x_max, y_min)]]

    rois = {'type': 'Feature',
            'id': 1,
            'properties': {'id': 1},
            'geometry': {'coordinates': window_polygon,
                         'type': 'Polygon'}}

    # Collect samples for the window
    if config.app.rasters_are_timeseries:
        timeseries = TimeSeries(rasters, rois, out_dir=None)
        timeseries.gather_samples(config=config, for_prediction=True)
        pixel_data = timeseries.get_samples(only_raster_data=True)
    else:
        samples = Samples(rasters, rois, out_dir=None)
        samples.gather_samples(config=config, for_prediction=True)
        pixel_data = samples.get_samples(only_raster_data=True)

    # Shape: rows=samples, columns=bands*timesteps
    data_array = pixel_data.to_numpy()

    # Find rows (samples) where any value is nan
    valid_rows = ~np.isnan(data_array).any(axis=1)

    # Reshape that to the window (valid mask)
    valid = valid_rows.reshape((shape[1], shape[2]))

    # Extract the valid rows for prediction
    data = data_array[valid_rows, :]

    return data, valid


def prediction(window: Window, rasters: List[Path], model_dict: dict,
               meta: dict, config: Configuration) -> None:
    """ Prediction for a window and save to a tmp file

    Args:
        window (rasterio.Window): A rasterio window
        rasters (List[Path]):  Rasters
        model_dict (dict): specifying the algorithm, model labels
        meta (dict): The meta files from source, destination and dest_proba
        config (Configuration)class): contains config
    """

    data, valid = \
        gather_data_for_prediction(rasters, meta, config, window)
    if not valid.any():
        return

    if not config.app.algorithm in \
            US_ALGORITHMS and config.app.algorithm != "knn_dtw":
        result = do_prediction_proba(
            data,
            valid,
            model_dict,
            config,
            meta,
            window
        )
    else:
        labels_external = model_dict['model'].predict(data)
        result = np.full(valid.shape, -9999, np.int32)
        result[valid] = labels_external

    # Write the classification results to tifs
    write_tifs(os.path.join(config.tmp_dir.name, 'classification'),
               window,
               meta['dst_meta'],
               result)


def predict_proba_single_class(
        data: np.ndarray, valid: np.ndarray, model_dict: dict,
        config: Configuration, meta: dict, window: Window) -> None:
    """Predict the probability for an array

    Args:
        data (np.ndarray): Features
        valid (np.ndarray): Valid values for prediction
        model_dict (dict): The model dictionary containing model and
        config (Configuration): The cli configuration
        meta (dict): rasterio meta for writing tif
        window: The rasterio Window for which to do the prediction

    Returns:
        None
    """
    meta.update(dtype=rasterio.dtypes.float32, nodata=-9999)
    result = np.full(valid.shape, -9999, np.float32)
    labels_internal_highest_p = model_dict['model'].decision_function(data)
    result_proba = result.copy().astype(np.float32)
    result_proba[valid] = labels_internal_highest_p

    write_tifs(os.path.join(config.tmp_dir.name, 'probability'),
               window,
               meta["dst_proba_meta"],
               result_proba)


def predict_single_class(
        data: np.ndarray, valid: np.ndarray, model_dict: dict,
        config: Configuration, meta: dict, window: Window) -> np.ndarray:
    """Predict the probability for an array

    Args:
        data (np.ndarray): Features
        valid (np.ndarray): Valid values for prediction
        model_dict (dict): The model dictionary containing model and
            algorithm info
        config (Configuration): The cli configuration
        meta (dict): rasterio meta for writing tif
        window: The rasterio Window for which to do the prediction

    Returns:
        predicted array

    """
    # Get the final classification of the highest probabilities
    predicted = model_dict['model'].predict(data)

    result = np.full(valid.shape, -9999, np.int32)

    # See if we need to output some of the modelled probabilities
    if config.supervised.probability:
        predict_proba_single_class(data,
                                   valid,
                                   model_dict,
                                   config,
                                   meta,
                                   window)
    result[valid] = predicted
    return result



"""Raster Utilities"""
# pylint: disable=cyclic-import
import itertools
import logging
import os
import re
import subprocess
import sys
from pathlib import Path
from typing import Any, List, Tuple

import numpy as np
import rasterio
from rasterio.windows import Window

from classifier.utils.config import Configuration

UTILS_RASTER_LOGGER = logging.getLogger(__name__)


def stitch_(out_ds: str, temp_dir: str,
            meta: dict, dtype: str = 'Int32') -> None:
    """Stitch all tifs together to a single tif

    Args:
        out_ds (str): output file
        temp_dir (str): The temp directory containing the tifs
        meta (dict): rasterio meta dictionary for output
        dtype (dtype): the raster's dtype

    """
    filelist = [os.path.join(temp_dir, x) for x in os.listdir(temp_dir) if
                x.endswith('tif')]
    input_list = temp_dir+'input_list.txt'
    with open(input_list, 'w', encoding='UTF-8') as tif_list:
        for files in filelist:
            tif_list.write(f'{files}\n')
    cmd_pred = (
        f"""gdalbuildvrt -srcnodata {meta['dst_meta']['nodata']} \
        {out_ds}.vrt \
        -q \
        -input_file_list {input_list}
        """
    )
    subprocess.call(cmd_pred, shell=True)
    cmd_pred = (
        f"""
        gdalwarp -ot {dtype} -q -multi -wo "NUM_THREADS=ALL_CPUS" -co \
        "TILED=YES"  -co "COMPRESS=DEFLATE" -co "BIGTIFF=YES" -overwrite \
        -srcnodata {meta['dst_meta']['nodata']} {out_ds}.vrt {out_ds}.tif
        """
    )
    subprocess.call(cmd_pred, shell=True)
    os.remove(out_ds+'.vrt')


def stitch(out_prefix: Path, meta: dict, config: Configuration) -> None:
    """Stitch the temporary files together

    Args:
        out_prefix (Path)): The output directory
        meta (dict): The rasterio meta dicts from source, destination and
                     dest_proba
        config (Configuration): contains config

    """
    # Find all directories that have chunks.
    dirs = list(x for x in os.listdir(config.tmp_dir.name))
    for directories in dirs:
        if directories.startswith('class'):
            out_file = out_prefix / 'classification'
            dtype = 'Int32'
        elif directories.startswith('probab'):
            out_file = out_prefix / 'probability'
            dtype = 'Float32'
        else:
            out_file = out_prefix / directories
            dtype = 'Float32'
        UTILS_RASTER_LOGGER.info("Now stitching %s", out_file)
        stitch_(out_file.as_posix(),
                os.path.join(config.tmp_dir.name, directories),
                meta,
                dtype)


def get_meta(rasters: List[Path],
             window_size: int = 1024) -> Tuple[List[Tuple[int, Window]],
                                               dict]:
    """ Get the metadata from the input file and create the windows list to
    iterate through

    Args:
        rasters (List[Path]): List of rasters
        window_size (int): The size of the windows that will be processed

    Returns:
        windows (List[int, Window]): List containing number of window and window
        dict: Containing info for rasterio write tiff
    """

    first_raster = rasters[0]
    # Read the metadata of the first and return metadata for dst and dst_proba
    with rasterio.open(first_raster, 'r') as first:
        f_meta = first.meta
        meta = f_meta.copy()
        meta.update(dtype=rasterio.dtypes.int32, nodata=-9999, count=1,
                    driver='GTiff')
        # Create the windows
        col_offs = np.arange(0, first.width, window_size)
        row_offs = np.arange(0, first.height, window_size)
        window_tuples = [x + (window_size, window_size) for x in list(
            itertools.product(col_offs, row_offs))]
        winlist = [rasterio.windows.Window(*x) for x in window_tuples]
        windows = list(zip(range(len(winlist)), winlist))
        meta_proba = meta.copy()
        meta_proba.update(dtype=rasterio.dtypes.float32, nodata=-9999)
    bandcount = []
    for raster in rasters:
        with rasterio.open(raster, 'r') as src_r:
            bandcount.append(src_r.count)
    return windows, {'f_meta': f_meta, 'dst_meta': meta,
                     'dst_proba_meta': meta_proba,
                     'bandcount': np.sum(bandcount)}


def open_single_raster(raster_path: Path, window: Window = None)\
        -> Tuple[np.ndarray, dict]:
    """Opens single raster and returns numpy array.

    Array shape: nbands x nrows x ncols.

    Args:
        raster_path(Path): location of the raster file
        window (rasterio.Window): subset window

    Returns:
        raster_array (np.ndarray): Raster data
        meta (dict): dictionary with raster metadata
    """

    with rasterio.open(raster_path, 'r') as raster:
        raster_array = raster.read(window=window)
        meta = raster.meta
        meta.update(
            width=raster_array.shape[2],
            height=raster_array.shape[1],
            transform=raster.window_transform(window))
        return raster_array, meta


def get_raster_date(raster: Path) -> Any:
    """Gets the date of a tif file from the name. The pattern that is expected
    is YYYY-MM-DD.

    If there are multiple dates with that format in the filename, only the
    first one will be used

    Args:
        raster (Path) : Path to raster file
    Returns:
        file_date (str): date of the raster in string format
    """
    file_date = re.findall(r'\d\d\d\d-\d\d-\d\d', raster.as_posix())[0]
    return file_date


def raster_warp_dst(raster: Path) -> dict:
    """Get the warp paramters of a raster:

    Args:
        raster(Path): path to a raster file

    Returns:
        warp_dst (dict): warp parameter dict
    """
    with rasterio.open(raster) as first:
        warp_dst = {
            "width": first.width,
            "height": first.height,
            "transform": first.transform,
            "crs": first.crs
        }
    return warp_dst


def count_bands(raster: Path) -> int:
    """Returns number of bands of a raster

    Args:
        raster (Path): path to raster file

    Returns:
        (int): Number of bands of a raster


    """
    with rasterio.open(raster, 'r') as src:
        return int(src.count)


def verify_and_count_bands(rasters: List[Path]) -> int:
    """Counts the number of bands for all the rasters and makes sure they all
    have the same number bands. Then it returns the number of bands that the
    rasters have

        Args:
            rasters(List[Path]): list of raster files

        Returns:
            b_count (int): number of bands in each raster

    """
    n_bands_list = [count_bands(raster) for raster in rasters]
    count = set(n_bands_list)
    if len(count) > 1:
        UTILS_RASTER_LOGGER.error("Number of bands per raster is not equal. "
                                  "Quitting...")
        sys.exit()
    return count.pop()


def get_bandnames(rasters: List[Path]) -> List[str]:
    """Get the filenames and add a number for the bands

    Args:
        rasters (List[Path]): The list of input rasters

    Returns:
        bandnames (List[str]): List of rasternames and bandsuffixes
    """
    bandnames = []
    for files in rasters:
        rastername = files.stem
        with rasterio.open(files) as src:
            nr_bands = src.count
            bandnames.extend([f'{rastername}_{x}' for x in
                              np.arange(1, nr_bands+1)])
    return bandnames



"""General utilities"""
import contextlib
import dataclasses
import inspect
import json
import logging
import os
import shutil
import sys
import tempfile
from itertools import product
from pathlib import Path
from typing import Any, Callable, List, Tuple, Optional, Dict, Union
from uuid import uuid4
from zipfile import ZipFile

import joblib
#from matplotlib.patches import Polygon
import numpy as np
import rasterio
from joblib import Parallel, delayed
from tqdm import tqdm
from shapely.geometry import shape, box

from classifier.settings import RASTER_EXTENSIONS, WORKSPACE
from classifier.utils.config import (
    Configuration, setup_config, PARAMETERS)

UTILS_GENERAL_LOGGER = logging.getLogger(__name__)


def get_available_model_args(model_args: dict, function: Callable) -> dict:
    """Gets the available arguments for the specific function
        Args:
            model_args (dict) : arguments to check
            function (func): function to check the kwargs from

        Returns:
            model_algorithm_args(list): model arguments that belong to function

    """
    kwarglist = inspect.getfullargspec(function)[0]
    model_algorithm_args = [x for x in model_args.keys() if x in kwarglist]
    return {k: model_args[k] for k in model_algorithm_args}


def progress(count: int, total: int, status: str = '') -> None:
    """A simple progress bar for the command line

    Args:
        count (int): Count between 0 and total
        total (int): Last iteration
        status (str): State

    """
    count += 1
    bar_len = 60
    filled_len = int(round(bar_len * count / float(total)))

    percents = round(100.0 * count / float(total), 1)
    p_bar = '=' * filled_len + '-' * (bar_len - filled_len)
    sys.stdout.write(f'[{p_bar}] {percents}% ...{status}\r')
    sys.stdout.flush()


def cli_init(
    output_name: str, rasters: List[str],
    overwrite: bool, config_location: Union[str, Path],
    rois: Optional[str] = None) \
        -> Tuple[Path, List[Path], Optional[Path], Configuration]:
    """Initialize everything before starting real work.

    Args:
        output_name (str): name of the output directory in workspace
        rasters (List[str]): Raster paths
        overwrite (bool): overwrite existing folder
        config_location (Path): location of the config file
        rois (str): rois path for input


    Returns:
        output_directory: path to output directory
        rasters (List[Path]): list of separate raster files
        rois (Path): path of rois workspace appended
        config (Configuration): dict containing Config

    """
    # Set Random Seed for reproducible results
    np.random.seed(0)

    if isinstance(config_location, str):
        config_location = WORKSPACE / config_location
    config = read_config(config_location)

    output_directory = create_output_dir(WORKSPACE, output_name, overwrite)
    # pylint: disable = consider-using-with
    config.tmp_dir = tempfile.TemporaryDirectory(dir=output_directory)
    init_logger(output_directory, config.app.log_level)
    params_str = config_as_str(config)
    UTILS_GENERAL_LOGGER.info(
        "\nRunning the Classifier with the following parameters:"
        "\n  %s", params_str)
    if rois is not None:
        rois = WORKSPACE / rois

    raster_paths = get_raster_paths(WORKSPACE, rasters)

    return output_directory, raster_paths, rois, config


def config_as_str(config: Configuration) -> str:
    """ Adds config values to string
    Args:
        config (Configuration) contains config

    Returns:
        params_str (str): parameters formatted in a string
    """
    params_str = ''
    for key, value in dataclasses.asdict(config).items():
        if isinstance(value, dict):
            params_str = params_str + f'{key}:\n'
            for key_2, value_2 in value.items():
                if isinstance(value_2, dict):
                    params_str = params_str + f'    {key_2}:\n'
                    for key_3, value_3 in value_2.items():
                        params_str = params_str + \
                            f'        {key_3}:  {value_3}\n'
                else:
                    params_str = params_str + f'    {key_2}:  {value_2}\n'
        else:
            params_str = params_str + f'{key}:  {value}\n'
    return params_str


def create_output_dir(workspace: Path, name: str, overwrite: bool) -> Path:
    """Create output directory.

    Args:
        workspace (Path): The workspace Path
        name (str): Location of the directory
        overwrite (bool): overwrite existing folder

    Returns:
        Output directory (Path) Path of output directory

    """
    if name is None:
        name = str(uuid4())[:6]
        UTILS_GENERAL_LOGGER.info(
            "No name argument found. Making new folder "
            "called: %s", name)
    output_directory = workspace / name
    if output_directory.exists():
        if overwrite:
            UTILS_GENERAL_LOGGER.warning("Overwriting existing directory!")
            shutil.rmtree(output_directory)
            os.mkdir(output_directory)
        else:
            UTILS_GENERAL_LOGGER.error(
                "Directory with name %s already exists. Either "
                "leave --name out,remove the directory %s, "
                "provide a unique name or turn on --overwrite.", name, name)
            sys.exit()
    else:
        os.mkdir(output_directory)
    return output_directory


def save_dict_as_json(out_file: Path, dict_to_save: dict) -> None:
    """Saves a dict to json

    Args:
        out_file (Path): file to write
        dict_to_save (dict): Dictionary to save

    """
    with open(out_file, 'w', encoding='UTF-8') as config:
        config.write(json.dumps(dict_to_save,
                                sort_keys=True,
                                indent=1,
                                default=str))


def save_config_as_json(**kwargs: Dict[str, Any]) -> None:
    """Save default config values.

    Save default values for classifier parameters in json in current workspace.

    Args:
        kwargs (dict): values to put into the config

    """
    out_file = Path(WORKSPACE) / 'config.json'
    dict_to_save = PARAMETERS
    # update dict with provided parameters
    dict_to_save = update_dict(dict_to_save, **kwargs)
    with open(out_file, 'w', encoding='UTF-8') as config:
        config.write(json.dumps(dict_to_save, indent=4, default=str))


def update_dict(dict_to_update: dict, **kwargs: Dict[str, Any]) -> dict:
    """Update dict with provided values

    Args:
        dict_to_update (dict): dict which values should be updated
        **kwargs (dict): keywords and values to update

    Returns:
        updated_dict (dict): updated dict
    """
    updated_dict = {}
    for key, value in dict_to_update.items():
        if isinstance(value, dict):
            updated_dict[key] = update_dict(value, **kwargs)
        else:
            updated_dict[key] = kwargs.get(key, value)
    return updated_dict


def read_config(config_path: Path) -> Configuration:
    """Read config parameter file and change defaults where necessary.

    Args:
        config_path (pathlib.Path): location of the config file

    Returns:
        config (Configuration): nested dataclass with all config values
    """
    config = setup_config(config_path)

    return config


def init_logger(output_directory: Path, log_level: str) -> None:
    """Set and initialize all logging info.

    Args:
        output_directory (Path): The location of the log file
        log_level (str)

    """
    logging.captureWarnings(True)
    logging.basicConfig(level=logging.DEBUG,
                        filename=output_directory / 'stdout.log')
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.DEBUG)

    # disable matplotlib logger
    logging.getLogger('matplotlib.font_manager').disabled = True

    # # create console handler and set level to info
    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setLevel(log_level)
    # # create formatter
    formatter = logging.Formatter(
        '%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    # # add formatter to ch
    console_handler.setFormatter(formatter)
    # # add Handlers to logger
    logging.getLogger('').addHandler(console_handler)


def write_tifs(
        temp_dir: str, window: rasterio.windows.Window, meta: dict,
        result: np.ndarray) -> None:
    """Writes the tifs of the individual windows

    Args:
        temp_dir: Temp dir location (str)
        window: rasterio window (tuple)
        meta: rasterio meta dictionary for output
        result: Result array

    Returns:
        Nothing

    """
    # Check output dir existence
    os.makedirs(temp_dir, exist_ok=True)
    meta['compress'] = 'deflate'
    with rasterio.open(
            os.path.join(temp_dir, f"c{window.col_off}_{window.row_off}.tif"),
            'w', **meta) as dst:
        dst.write_band(1, result)


def save_model(model_dict: dict, out_dir: Path, config: Configuration) -> None:
    """Save the model as a pickle  and metadata  as a json and zip

    Args:
        model_dict (dict):  Contains the model as well as metadata
        out_dir (Path):      Path where to write model file and meta
        config (Configuration): Configuration parameters

    """
    to_save = model_dict.copy()
    meta_tmp = tempfile.mktemp()
    joblib_tmp = tempfile.mktemp()

    # Take the model out of the dictionairy
    model = to_save.pop('model')

    # Save the model as pickle
    with open(joblib_tmp, 'wb') as model_file:
        joblib.dump(model, model_file, compress=3)

    # Save meta as json
    save_dict_as_json(Path(meta_tmp), to_save)
    to_write = {
        meta_tmp: Path(f"{config.name}_meta.json"),
        joblib_tmp: Path(f"{config.name}.model")
    }

    # Zip to output directory
    zipfile_path = out_dir / 'model.zip'
    write_zipfile(to_write, zipfile_path)


def write_zipfile(files_to_write: dict, zipfile_path: Path) -> None:
    """Writes files to a zipfile

    Args:
        files_to_write (dict): dictionary of path, name of the files to
                               write to the zipfile (e.g.
                               {
                                  '/workspace/weirdname.json':
                                  'nice_name.json'
                               }
        zipfile_path (Path): Path to where the zipfile goes
    """
    with ZipFile(zipfile_path, 'w') as model_zip:
        for files in files_to_write:
            print(files, files_to_write[files])
            model_zip.write(files, files_to_write[files])


def get_raster_paths(workspace: Path, rasters: List[str]) -> List[Path]:
    """Get full paths for all rasters.

    Args:
        workspace (Path): The workspace Path
        rasters (List[Path]): List of rasters from the cli

    Returns:
        List[Path]: rasters with their full paths

    """
    paths = []

    for raster in rasters:
        path = workspace / raster
        if path.is_dir():
            paths += [path / r for r in sorted(path.iterdir())]
        else:
            paths.append(path)
    found_rasters = [path for path in paths if path.suffix in RASTER_EXTENSIONS]
    if not found_rasters:
        UTILS_GENERAL_LOGGER.info(
            "Found no rasters.")
        sys.exit(1)
    return found_rasters


def unzip_model_file(model_file: Path) -> Tuple[Path, Path]:
    """Unzips a model file and returns paths to metadata and model pickle

    Args:
        model_file (Path): path to the model zipfile

    Returns:
        meta_file (Path): path to tmp metafile
        pickle_file (Path): path to tmp pickle file
    """
    tmpdir = Path(tempfile.mkdtemp())
    with ZipFile(model_file, 'r') as zipf:
        zipf.extractall(tmpdir)

    model_files = list(tmpdir.iterdir())
    meta_file = tmpdir / [x for x in model_files if x.suffix == '.json'][0]
    pickle_file = tmpdir / [x for x in model_files if x.suffix == '.model'][0]
    return meta_file, pickle_file


def read_model(model_file: Path) -> dict:
    """Function to read the model file and return a pickle

    Args:
        path: The path to the model zipfile

    Returns:
        a dictionary containing a model, and metadata

    """
    # Unzip
    meta_file, joblib_file = unzip_model_file(model_file)
    # Open the metadata and print the modeltypes
    with open(meta_file, encoding='UTF-8') as json_config:
        model_dict = json.load(json_config)

    params_str = '\n'.join(
        [f'{key} :  {value}' for key, value in sorted(model_dict.items())])

    with open(joblib_file, 'rb') as m_file:
        model_dict['model'] = joblib.load(m_file)

    UTILS_GENERAL_LOGGER.info(
        "\nLoaded a saved model, with the following metadata:"
        "\n  %s", params_str)

    return model_dict


def parallel_function(func: Callable,
                      iterable: List[dict],
                      ncpus: int = 1) -> Any:
    """Runs a function in parallel using Joblib

    Args:
        func: Function to run
        iterable (list): List of kwargs for the function
        ncpus (int): number of cpus to use

    Returns:
        list of results
    """
    with tqdm_joblib(tqdm(total=len(iterable))) as _:
        result = Parallel(n_jobs=ncpus)(delayed(func)(**x) for x in iterable)

    return result


@ contextlib.contextmanager
def tqdm_joblib(tqdm_object: tqdm) -> None:
    """Patch joblib to report a tqdm progress bar for many thread jobs

    Args:
        tqdm_object (tqdm.tqdm): Iterable decorated with progress bar

    Yields:
        tqdm.tqdm: Iterable decorated with progress bar
    """

    # pylint: disable=too-few-public-methods
    class TqdmBatchCompletionCallback(joblib.parallel.BatchCompletionCallBack):
        """ Class to pass completion callback to tqdm"""

        def __call__(self, out: object) -> Any:
            """Updates tqdm and returns joblib call

            Args:
                 out (object): Any object

            Returns:
                object: Any object
            """
            tqdm_object.update(n=self.batch_size)
            return super().__call__(out)

    old_batch_callback = joblib.parallel.BatchCompletionCallBack
    joblib.parallel.BatchCompletionCallBack = TqdmBatchCompletionCallback
    try:
        yield tqdm_object
    finally:
        joblib.parallel.BatchCompletionCallBack = old_batch_callback
        tqdm_object.close()


def rois_in_raster_extent(rasters: List[Path], rois: List[dict]) -> bool:
    """Check if the rois lie in raster extent

    Args:
        rasters (List[Path]): List of raster paths
        rois (List[dict]): List of rois path

    Returns:
        bool: True if all rois lay in raster
    """
    all_rois_in_raster = True
    with rasterio.open(rasters[0]) as src:
        raster_geom = box(*src.bounds)
        for roi in rois:
            roi_geom = shape(roi['geometry'])
            if not raster_geom.contains(roi_geom):
                UTILS_GENERAL_LOGGER.warning(
                    "Roi %s does not lie in the raster", str(roi['id'])
                )
                all_rois_in_raster = False
    return all_rois_in_raster


def dict_product(d: dict) -> dict:  # pylint: disable=invalid-name
    """ Function to build all combinations of multiple lists. But
    returns a dict with the argument name as key to still know which
    value belongs to which key (element)"""
    keys = d.keys()
    for element in product(*d.values()):
        yield dict(zip(keys, element))



"""Settings file for Classifier"""
from pathlib import Path

from sklearn.cluster import KMeans, MiniBatchKMeans

from classifier.models import RandomForest, SingleClass, XGBoost, KnnDtw
WORKSPACE = Path('/workspace/')

RASTER_EXTENSIONS = frozenset(['.tif', '.vrt', '.jp2'])

US_ALGORITHMS = ['us_kmeans', 'us_kmeans_minibatch']

# ##----THE ALGORITHMS---###
ALGORITHMS = [
    "randomforest", "xgboost", "singleclass", "us_kmeans",
    "us_kmeans_minibatch", "knn_dtw"
]
CLASSIFIERS = [RandomForest, XGBoost, SingleClass, KMeans,
               MiniBatchKMeans, KnnDtw]
ALGORITHM_DICT = dict(zip(ALGORITHMS, CLASSIFIERS))



"""An sk-learn classifier for landcover classification"""

import logging
import sys
import time
from pathlib import Path
from typing import List, Optional

from classifier.accuracy import (plot_feature_importances,
                                 write_confusion_matrix)
from classifier.predict import prediction
from classifier.settings import US_ALGORITHMS
from classifier.train import train_dataset, train_kmeans
from classifier.utils.config import Configuration
from classifier.utils.general import parallel_function, read_model, save_model
from classifier.utils.raster import get_meta, stitch
from classifier.samples import Samples, TimeSeries
from classifier.utils.vector import get_rois_extent


START = time.time()
CLASSIFIER_LOGGER = logging.getLogger(__name__)


def train(samples: Samples, out_dir: Path, config: Configuration,
          rois_extent: List[float] = None) -> dict:
    """Train a model using samples and rasters

    Args:
        samples (Samples): Samples dataframe containing pixel
            values and class values
        rasters (List[Path]): A list of raster files
        out_dir (Path): Path where to write the dataset file
        config (Configuration): contains config
        rois_extent (List[float]): The extent of the training data

    Returns:
        model_dict (dict): Containing the name, model and label
                           encoder.
    """
    CLASSIFIER_LOGGER.info("\n####-----Training----#####\n")
    #windows, _ = get_meta(rasters, config.app.window)
    if config.app.algorithm in US_ALGORITHMS:
        # Do unsupervised
        array = samples.get_samples()
        model_dict = train_kmeans(array,
                                  config)
    else:  # All supervised methods
        model_dict, test = train_dataset(
            samples,
            out_dir,
            config
        )
        if config.app.algorithm in ['randomforest', 'xgboost'] \
                and config.accuracy.perform_assesment:
            CLASSIFIER_LOGGER.info(
                "\n####----Accuracy Assessment----#####\n")
            # Do the accuracy analysis
            cm_fn = out_dir / 'confusion_matrix'
            fi_fn = out_dir / 'feature_importance.png'
            write_confusion_matrix(model_dict, test, cm_fn)
            plot_feature_importances(model_dict, fi_fn)
    # save the model as a python pickle
    if not rois_extent is None:
        model_dict['rois_bounds'] = rois_extent
    if config.app.model_save:
        save_model(model_dict, out_dir, config)
    CLASSIFIER_LOGGER.info("\nFinished Training\n")
    return model_dict


def predict(
        model_dict: dict, rasters: List[Path],
        out_dir: Path, config: Configuration) -> None:
    """Prediction function using a trained model and raster files

    Args:
        model_dict (dict): Containing the name, model and label
                        encoder.
        rasters (List[Path]): A list of raster files
        out_dir (Path): Path where to write the dataset file
        config (Configuration): contains config
    """
    threads = config.app.threads
    CLASSIFIER_LOGGER.info("\n####-----Prediction----#####\n")
    windows, meta = get_meta(rasters, config.app.window)

    # Set model n_jobs to 1 for prediction
    if 'n_jobs' in model_dict['model'].get_params():
        model_dict['model'] = \
            model_dict['model'].set_params(**{'n_jobs': 1})

    iterable = [
        {
            'window': window[1],
            'rasters': rasters,
            'model_dict': model_dict,
            'meta': meta,
            'config': config
        }
        for window in windows
    ]
    if threads > 1 or threads == -1:
        parallel_function(prediction,
                          iterable,
                          threads)
    else:
        for wins in iterable:
            prediction(**wins)

    # ##--------------------STITCHING-----------------------------------###

    CLASSIFIER_LOGGER.info("\n####-----Stitching----#####\n")

    # Run the gdalwarp command in the command line
    stitch(out_dir, meta, config)
    CLASSIFIER_LOGGER.info("Cleaning up..")
    config.tmp_dir.cleanup()
    CLASSIFIER_LOGGER.info(
        "Total run time was %i seconds", (int(time.time() - START)))


def check_input_choices(
        config: Configuration, rois: Optional[Path] = None) -> None:
    """Checks if the chosen options and provided data match

    Args:
        rasters (List[Path]): A list of raster files
        config (Configuration): contains config
        rois (Path): The path to a rois file

    Returns:
        None
    """

    # If rois and samples are None, only unsupervised is possible
    if rois is None and config.app.samples is None:
        if config.app.algorithm not in US_ALGORITHMS:
            CLASSIFIER_LOGGER.error(
                "You have chosen a supervised classification method but didn't "
                "provide rois or samples. Either provide one of these or run "
                "an unsupervised classification.")
            sys.exit(1)

    if rois is not None and config.app.samples is not None:
        CLASSIFIER_LOGGER.error(
            "You provided rois and samples but not both can be used. "
            "Only provide one of them.")
        sys.exit(1)

    if rois is not None or config.app.samples is not None:
        if config.app.algorithm in US_ALGORITHMS:
            CLASSIFIER_LOGGER.error(
                "You have chosen an unsupervised classification "
                "method but provided samples or rois. Please choose "
                "a supervised method to use them or don't provide them for "
                "an unsupervised classification.")
            sys.exit(1)


def classify_(
        rasters: List[Path], config: Configuration,
        out_dir: Path, rois: Optional[Path] = None) -> None:
    """Entry point function for pixel-based classification

        Args:
            rasters (List[Path]): A list of raster files
            config (Configuration): contains config
            out_dir (Path): Path where to write the dataset file
            rois (Path): The path to a rois file

    """

    CLASSIFIER_LOGGER.info("Starting Classification...")
    if config.app.model is not None:
        CLASSIFIER_LOGGER.info(
            "You provided a model. Remember that this model has to "
            "be already trained and provided rois/samples will "
            "not be used. Only predictions are done."
        )
        # Model provided, only do prediction
        model_dict = read_model(config.app.model)
    else:
        CLASSIFIER_LOGGER.info(
            "No model provided. Train a new one."
        )

        check_input_choices(config, rois)

        # gather samples
        samples = Samples(rasters, rois, out_dir)
        samples.gather_samples(config)

        rois_extent = []
        if rois is not None:
            rois_extent = get_rois_extent(rois)

        # train
        model_dict = train(samples, out_dir,
                           config, rois_extent)

    predict(model_dict, rasters, out_dir, config)


def classify_timeseries(rasters: List[Path],
                        config: Configuration,
                        out_dir: Path,
                        rois: Path) -> None:
    """Main function to classifiy timeseries.

    For now, only from start to finish is supported. ie, supplying a model
    or samples does not work.

        Args:
            rasters (List[Path]): List of raster paths
            config (Configuration): contains config
            out_dir (Path): Path to output directory
            rois (Path): Path to rois file

    """
    CLASSIFIER_LOGGER.info("Timeseries Classification")

    if config.app.algorithm == 'unsupervised':
        CLASSIFIER_LOGGER.error(
            "Unsupervised classification not supported yet")
        sys.exit()

    if config.app.model is not None:
        # Model provided, only do prediction
        model_dict = read_model(config.app.model)
    else:
        # No Model provided
        # gather samples
        timeseries = TimeSeries(rasters, rois, out_dir)
        timeseries.gather_samples(config)

        model_dict = train(timeseries,
                           out_dir,
                           config,
                           rois_extent=None)

    predict(model_dict, rasters, out_dir, config)


if __name__ == "__main__":
    pass



"Samples class with child Timeseries class"
import datetime
import logging
import os
import sys
from pathlib import Path
from typing import Any, List, Optional, Union, Dict

import fiona
import numpy as np
import pandas as pd
import rasterio
from rasterio import mask
from rasterio.enums import Resampling
from rasterio.vrt import WarpedVRT
from rasterio.windows import Window
from sklearn.ensemble import IsolationForest
from sklearn.impute import SimpleImputer

from classifier.dataprep import class_counts
from classifier.settings import US_ALGORITHMS
from classifier.utils.config import Configuration
from classifier.utils.general import parallel_function, rois_in_raster_extent
from classifier.utils.raster import (count_bands, get_bandnames, get_meta,
                                     get_raster_date, raster_warp_dst,
                                     verify_and_count_bands)

SAMPLES_LOGGER = logging.getLogger(__name__)


class Samples():
    """
    Class that holds all the samples related data

    Attributes
    ----------
    rasters : List[Path]
        raster paths
    rois : List[dict]
        contains the rois each consisting of one dict respectively
    out_dir : Path
        main output directory for result files and co.
    samples : pd.DataFrame
        contains the sampled data

    """

    def __init__(self,
                 rasters: List[Path],
                 rois: Union[Dict[str, object],
                             Optional[Path]],
                 out_dir: Path) -> None:

        self.rasters: List[Path] = rasters
        self.rois: List[dict] = None

        if rois is not None:
            self.rois = []
            if isinstance(rois, dict):
                # prediction
                self.rois.append(rois)
            else:
                # training
                with fiona.open(rois, "r") as shapefile:
                    # unpack the rois from shapefile in list
                    for roi in shapefile:
                        self.rois.append(roi)
        self.out_dir = out_dir
        self.samples = None

        if rasters is not None and rois is not None:
            self.check_geo_input_data(self.rasters, self.rois)

    @staticmethod
    def read_samples(samples_file: Path) -> Any:
        """Read a samples file.

        Args:
            samples_file (Path): CSV or Pickle file containing samples

        Returns:
            sample_df (pd.DataFrame): DF containing samples

        """
        SAMPLES_LOGGER.info("Read samples from file...")
        if samples_file.suffix == '.pkl':
            return pd.read_pickle(samples_file)
        if samples_file.suffix == '.csv':
            return pd.read_csv(samples_file, index_col=0)
        SAMPLES_LOGGER.error(
            "Unknown samples file type")
        return None

    def write_samples(self, file_type: str = 'csv') -> None:
        """Write samples to pickle or csv

        Args:
            file_type (str): File type: csv or pkl

        Returns:
            None

        """
        SAMPLES_LOGGER.info("Save samples to file.")
        if self.samples is not None:
            if file_type == 'csv':
                self.samples.to_csv(
                    os.path.join(self.out_dir, 'samples.csv')
                )
            elif file_type == 'pkl':
                self.samples.to_pickle(
                    os.path.join(self.out_dir, 'samples.pkl')
                )
            else:
                SAMPLES_LOGGER.info(
                    "Unknown file type. Choose csv or pkl")
        else:
            SAMPLES_LOGGER.info(
                "No Samples collected yet.")

    def get_samples(self, only_raster_data: bool = False) -> pd.DataFrame:
        """Returns the samples. If only raster_data true, return only the
        raster values and no class/roi_fid

        Args:
            only_raster_data (bool, optional): Exclude class and roi_fid column
            from returned dataframe

        Returns:
            pd.DataFrame: samples dataframe
        """
        if only_raster_data:
            return self.samples.loc[
                :, ~self.samples.columns.get_level_values(0).isin(
                    ['class', 'roi_fid'])]
        return self.samples

    def get_labels(self) -> pd.Series:
        """Return only the labels

        Returns:
            pd.Series: contains class labels
        """
        if self.samples is not None:
            return self.samples['class']
        return None

    def gather_samples(self,
                       config: Configuration,
                       for_prediction: bool = False) -> None:
        """Gets the samples

        Either from a provided sample file or it will collect the samples
        from provided rasters and rois

        Args:
            config (Configuration): contains config
            for_prediction (bool): Do not read samples from file
                or save samples to file if for_prediction

        Returns:
            None
        """
        # Gather samples
        if config.app.samples is not None and not for_prediction:
            # samples file provided
            self.samples = self.read_samples(
                config.app.samples)
        elif self.rois is not None:
            # gather samples from rasters and rois
            self.samples = self.create_samples(
                self.rasters,
                self.rois,
                config,
                for_prediction)
            if self.out_dir is not None and not for_prediction:
                self.write_samples()
        elif config.app.algorithm in US_ALGORITHMS:
            self.samples = self.create_samples_for_unsupervised(
                self.rasters, config)
        else:
            self.samples = None

    def create_samples(
            self, rasters: List[Path],
            rois: List[dict], config: Configuration,
            for_prediction: bool) -> pd.DataFrame:
        """Create the entire dataset based on rasters and rois

        Args:
            rasters (List[Path]): list of raster file names
            rois (List[Dict]): Contains each roi as a single dict
            config (Configuration): contains config
            for_prediction (bool): changes imputation/samples deleting behaviour

        Returns:
            samples (pd.DataFrame): dataset with samples
        """

        # Gather the samples and return a full df of all samples
        bandnames = get_bandnames(rasters)
        samples = self.gather_samples_from_data(rasters,
                                                rois,
                                                bandnames,
                                                None,
                                                config,
                                                for_prediction
                                                )
        return samples

    def fill_missing_data(
            self, samples: pd.DataFrame, config: Configuration,
            for_prediction: bool = False) -> pd.DataFrame:
        """ Handles the data holes by transforming infinity values to nan and
        then either filling all nan values with the chosen imputation method
        or deleting them.

        Args:
            samples (pd.DataFrame): contains the samples
            config (Configuration): Configuration
            for_prediction (bool): if its for prediction, do not drop NaNs

        Returns:
            samples (pd.DataFrame): dataframe with handled holes
        """

        # Check for numeric values
        samples, contains_nans = self.check_data(samples)

        # Imputation
        if contains_nans and config.app.rasters_are_timeseries:
            # If time series, impute through time
            return self.impute_ts(samples)
        if contains_nans and config.app.imputation:
            return self.impute(samples, config)
        if contains_nans and not for_prediction:
            # delete empty samples
            SAMPLES_LOGGER.warning("""The sampled dataset contains nan's.
            Imputation is set to False so all rows containing NaNs will be
            deleted. If you want to prevent this from happening, please set the
            app_imputation parameter to True and choose a strategy and constant
            if necessary.""")
            return samples.dropna(how='any')
        return samples

    def impute(self, samples: pd.DataFrame,
               config: Configuration) -> pd.DataFrame:
        """Impute a single raster

        Args:
            samples (pd.DataFrame): dataframe with samples
            config (Configuration): Configuration dataclass

        Returns:
            pd.DataFrame: imputed samples dataframe
        """
        if config.app.imputation_strategy == 'interpolate':
            data_filled_df = samples.interpolate(
                method="linear", axis=0, limit_direction="both")
        else:
            # Other imputation methods
            # Check if a feature column is None => will get deleted by
            # sklearn imputer (cant get imputed)
            if samples.isnull().all().any():
                SAMPLES_LOGGER.warning(
                    "%s Features have no values, cant be imputed.",
                    samples.isnull().all().sum())
                return samples
            data_filled = self.impute_values(samples, config)
            data_filled_df = pd.DataFrame(
                data_filled, index=samples.index, columns=samples.columns)
        return data_filled_df

    @staticmethod
    def check_data(dataset: pd.DataFrame) -> Union[pd.DataFrame, bool]:
        """Checks data for invalid values like np.inf and np.nan

        Args:
            dataset (pd.DataFrame): samples dataframe

        Returns:
            Union[pd.DataFrame, bool]: samples dataframe and if it
                contains nan values
        """
        # Replaced infinity values
        dataset = dataset.replace([np.inf, -np.inf], np.nan)

        # Check for nans
        contains_nans = dataset.isnull().values.any()
        return dataset, contains_nans

    def gather_samples_from_data(
            self,
            rasters: List[Path],
            rois: List[dict],
            bands: List[str],
            dates: List[datetime.datetime],
            config: Configuration,
            for_prediction: bool = False) -> pd.DataFrame:
        """Gathers the pixel values for all the rasters and
        combines them in a df

        The DataFrame is a multiColumn dataframe where
        column level 0 is the class

            Args:
                rasters (List[Path]): containg raster paths
                rois (List[dict]): containing the roi geometry dictionaries
                bands (List[str]): Contains band names
                dates (List[datetime.datetime]): Contains dates
                config (Configuration): contains config
                for_prediction (bool): Doesn't delete samples if
                    for prediction

            returns:
                samples_df (pd.DataFrame): DataFrame with of samples
        """
        # warp dst from first raster
        warp_dst = raster_warp_dst(rasters[0])

        # Parallel gathering of samples from the rois and rasters
        if rois is not None:
            kwargs = [
                {
                    'roi': roi,
                    'rasters': rasters,
                    'config': config,
                    'warp_dst': warp_dst,
                    'dates': dates,
                    'bands': bands,
                    'for_prediction': for_prediction
                }
                for roi in rois
            ]
        gathered_samples_list = parallel_function(
            self._gather_samples_per_roi,
            kwargs,
            ncpus=config.app.threads
        )
        if gathered_samples_list:
            # Concat results from all the rois
            samples_df = pd.concat(gathered_samples_list, axis=0)
        else:
            SAMPLES_LOGGER.error(
                "Couldn't collect any samples from any roi."
                "Check your rois or rasters."
            )
            sys.exit(1)

        # make class column instead of index
        samples_df.reset_index(level='class', inplace=True)

        # Remove outliers (right now only for non timeseries)
        if not for_prediction:
            if config.supervised.remove_outliers \
                    and not config.app.rasters_are_timeseries:
                samples_df = self.outlier_removal(samples_df)

        # make roi_fid a column instead of index
        samples_df.reset_index(level='roi_fid', inplace=True)

        # Reindex samples id
        samples_df.reset_index(inplace=True, drop=True)

        return samples_df

    def _gather_samples_per_roi(
            self, roi: dict,
            rasters: List[Path],
            config: Configuration,
            warp_dst: dict,
            dates: List[datetime.datetime],
            bands: List[str],
            for_prediction: bool) -> Union[None, pd.DataFrame]:
        """Gather samples for a roi, for usage in MP function
        Creates a dataframe with rows=sample and columns=bands

            Args:
                roi (dict shapely.geometry): region of interest
                rasters (List[Path]): raster list
                config (dataclass): contains config
                warp_dst (dict): warp parameters for raster
                df_index (list): index to use for creation of samples df
                bands (list): bands to use from the tifs
                for_prediction (bool): changes samples imputing/deleting
                    behaviour
        """
        # Gather all the samples for this roi from all rasters
        roi_samples = self._gather_samples_for_roi(
            roi,
            rasters,
            warp_dst
        )
        if roi_samples is not None:

            # Samples count for roi
            sample_count = roi_samples.shape[0]

            # Extract roi id (class)
            roi_class = int(roi['properties']['id'])
            roi_classes = sample_count * [roi_class]

            # Extract fid (unique id of each roi)
            roi_fid = int(roi['id'])
            roi_fids = sample_count * [roi_fid]

            # Sample number
            sample_id = np.arange(sample_count)

            # Column Index
            if dates:
                # Save also date if it's a timeseries
                columns = pd.MultiIndex.from_product(
                    [dates, bands],
                    names=['Date', 'Band']
                )
            else:
                columns = bands

            # Row Index
            index = pd.MultiIndex.from_tuples(
                list(zip(roi_classes, roi_fids, sample_id)),
                names=['class', 'roi_fid', 'pixel']
            )

            # Samples DataFrame for roi
            # Shape: rows=samples, columns=dates(if timeseries) * bands
            roi_samples_df = pd.DataFrame(
                roi_samples,
                index=index,
                columns=columns)

            # Handle invalid data
            roi_samples_df = self.fill_missing_data(
                roi_samples_df,
                config,
                for_prediction)

            if roi_samples_df.empty:
                return None

            # Samples which still have data gaps after imputation, are deleted
            # if its for training (not deleted for prediction, will get masked
            # out later)
            if not for_prediction:
                if roi_samples_df.isnull().any(axis=1).any():
                    # Sample seems to have no values for
                    # a given band for any raster timestep
                    SAMPLES_LOGGER.warning(
                        "Deleting samples from roi %s. "
                        "Could not find any values for them. ", roi_fid)
                    roi_samples_df = roi_samples_df.dropna()
            return roi_samples_df
        return None

    @staticmethod
    def _gather_samples_for_roi(roi: dict, rasters: List[Path],
                                warp_dst: dict) -> np.ndarray:
        """Get sample values from rasters warped to
        the specified warp destination

            Args:
                roi (fiona shape): region of interest
                rasters (List[Path]): list of raster file names
                warp_dst (dict): warp parameters for raster

            Returns:
                np.array [nsamples, nfeatures] of samples from within rois.
        """

        try:
            roi_values_list = []
            for files in rasters:
                # For each raster extract all band values.
                # Shape: rows=bands, columns=samples
                with rasterio.open(files) as src:
                    with WarpedVRT(
                        src,
                        resampling=Resampling.nearest,
                        width=warp_dst["width"],
                        height=warp_dst["height"],
                        transform=warp_dst["transform"],
                        crs=warp_dst["crs"]
                    ) as vrt:
                        selection, _, window = mask.raster_geometry_mask(
                            vrt,
                            [roi["geometry"]],
                            crop=True)
                        raster_subset = vrt.read(window=window)

                # rasterio convention: outside shape=True, inside=False.
                # We invert.
                samples_per_raster = raster_subset[:, ~selection]
                roi_values_list.append(samples_per_raster)
            # Concatenate the roi values from all rasters along rows.
            # Then transpose. Shape: rows=samples, columns=bands*timesteps
            roi_values = np.concatenate(roi_values_list, axis=0).T
        except ValueError:
            SAMPLES_LOGGER.debug("ROI %s OUT OF BOUNDS.. Continuing without "
                                 "it", roi['geometry']['coordinates'][0][0])
            roi_values = None
        return roi_values

    @staticmethod
    def impute_values(
            dataset: pd.DataFrame, config: Configuration) -> np.ndarray:
        """Impute values

        Uses the sklearn SimpleImputer to impute missing values.

        Args:
            dataset (pd.DataFrame): DataFrame containing Nans
            config (Configuration): Parameters to use for imputation

        Returns:
            dataset (np.ndarray): DataFrame with Nans imputed
        """

        verbosity_level = 0 if config.app.log_level == 'INFO' else 2

        # sklearn Imputer
        imputer = SimpleImputer(strategy=config.app.imputation_strategy,
                                fill_value=config.app.imputation_constant,
                                verbose=verbosity_level)
        # imputer expects ndarray
        return imputer.fit_transform(dataset.values)

    @staticmethod
    def outlier_removal(samples: pd.DataFrame) -> pd.DataFrame:
        """Outlier removal from samples using Isolation Forest

        Args:
            samples (pd.DataFrame): samples with possible outliers

        Returns:
            samples (pd.DataFrame): samples with filtered outliers
        """
        SAMPLES_LOGGER.info("Now checking for outliers")
        cols = [x for x in samples.columns if x != 'class']
        if_model = IsolationForest(contamination='auto')

        samples['inlier'] = 0
        lulc_classes = samples['class'].unique()

        for lulc_class in lulc_classes:
            samples.loc[samples['class'] == lulc_class, 'inlier'] = \
                if_model.fit_predict(samples[samples['class'] == lulc_class][
                    cols])

        # get the number of removed pixels and give output
        class_counts(samples)
        cols_to_return = [x for x in samples.columns if not x == 'inlier']
        return samples[samples['inlier'] > 0][cols_to_return]

    def create_samples_for_unsupervised(self, rasters: List[Path],
                                        config: Configuration) -> np.ndarray:
        """Create an input array for unsupervised classification

        Function for creating the input array for the training of the
        unsupervised models. A train_size relative to the entire dataset can be
        set to reduce the amount of memory needed

        Args:
            rasters (List[Path]): Input rasters
            config (Configuration): Contains config

        Returns:
            data_array (np.ndarray): A data array to use in model training"""

        SAMPLES_LOGGER.info(
            "Gathering random samples from rasters for "
            "unsupervised classification")
        warp_dst = raster_warp_dst(rasters[0])
        windows, _ = get_meta(rasters, config.app.window)

        # create_rois_for_us(windows, config)

        subset_windows = np.random.choice(
            np.array(windows)[:, 1],
            size=int(len(windows) * config.unsupervised.trainfraction),
            replace=False)
        n_windows = len(subset_windows)
        if n_windows == 0:
            SAMPLES_LOGGER.error(
                "Not enough subset data."
                "Please increase -us_train_size parameter"
            )
            sys.exit(1)
        SAMPLES_LOGGER.info("\nUsing %i windows for training", n_windows)

        kwargs = [{
            'rasters': rasters,
            'warp_dst': warp_dst,
            'window': window
        } for window in subset_windows]

        gathered_us_data = parallel_function(
            self.get_window_data_for_us_train,
            kwargs,
            ncpus=config.app.threads
        )

        data_array = np.concatenate(gathered_us_data, axis=0)

        dataset = pd.DataFrame(data_array)

        dataset = self.fill_missing_data(dataset, config)
        SAMPLES_LOGGER.info("\nCollected data from windows")
        return dataset

    @staticmethod
    def get_window_data_for_us_train(
            rasters: List[Path],
            warp_dst: dict,
            window: Window) -> np.ndarray:
        """Gets the pixel values for window and adds it to an MP list

        Args:
            rasters (List[Path]): List of rasters to get data from
            warp_dst (dict): warp parameters for raster
            window (rasterio.Window): Window to process

        """
        data_values = []
        # Create a vrt with the window and add the data to the array
        for raster in rasters:
            with rasterio.open(raster, 'r') as src:
                with WarpedVRT(src,
                               resampling=Resampling.nearest,
                               width=warp_dst["width"],
                               height=warp_dst["height"],
                               transform=warp_dst["transform"],
                               crs=warp_dst["crs"],
                               ) as vrt:
                    data_values.append(vrt.read(window=window))
        old_shape = np.shape(data_values)  # (1, channels, x, y)
        data_array = np.reshape(
            data_values,
            (old_shape[0] * old_shape[1],
             old_shape[2] * old_shape[3])).transpose()
        return data_array

    @staticmethod
    def check_geo_input_data(
            rasters: List[Path], rois: List[dict]) -> bool:
        """Check if all the rois are inside the rasters extent

        Args:
            rasters (List[Path]): Raster Paths
            rois (List[dict]): List of roi dict

        Returns:
            bool: True if all rois lie inside the rasters extent
        """

        # only do this for training (more than 1 rois) as windows for
        # prediction sometimes lie partially outside
        if rois is not None and len(rois) > 1:
            rois_lie_inside_rasters = rois_in_raster_extent(
                rasters, rois)
            return rois_lie_inside_rasters
        return True


class TimeSeries(Samples):
    """Class for Timeseries samples, subclass of Sample

    Overwrites some methods of Sample which need to be
    adjusted for time series
    """

    def __init__(self,
                 rasters: List[Path],
                 rois: Union[Dict[str, object],
                             Optional[Path]],
                 out_dir: Path) -> None:
        super().__init__(rasters, rois, out_dir)

        if rasters is not None:
            self.check_geo_input_data_ts(self.rasters)

    @ staticmethod
    def read_samples(samples_file: Path) -> pd.DataFrame:
        """Read a timeseries samples file.

        Args:
            samples(path): file containing samples

        Returns:
            sample_df(pd.DataFrame): Contains samples

        """
        if samples_file.suffix == ".pkl":
            SAMPLES_LOGGER.info("Load Pickle sample dataset\n")
            return pd.read_pickle(samples_file)
        SAMPLES_LOGGER.error(
            "\nFile format of sample file not supported.\n"
            "Supported samples files are Pickle: *.pkl")
        return None

    def write_samples(self, file_type: str = 'pkl') -> None:
        """Write samples to pickle or csv

        Args:
            file_type (str): File type: csv or pkl

        Returns:
            None

        """
        SAMPLES_LOGGER.info("Save Samples as Pickle...")
        if self.samples is not None:
            if file_type == 'pkl':
                self.samples.to_pickle(
                    os.path.join(self.out_dir, 'samples_ts.pkl')
                )
            else:
                SAMPLES_LOGGER.info(
                    "Other save file format currently not supported")
        else:
            SAMPLES_LOGGER.info(
                "No Samples collected yet.")

    def gather_samples(
            self, config: Configuration, for_prediction: bool = False) -> None:
        """Samples for time series from a list of rasters and a polygon file

            Args:
                config (Configuration): contains config
                for_prediction (bool): Do not read samples from file
                or save samples to file if for_prediction

            Returns:
                None
        """
        # Gather samples
        if config.app.samples is not None and not for_prediction:
            # samples file provided
            self.samples = self.read_samples(
                config.app.samples)
        elif self.rois is not None:
            # gather samples from rasters and rois
            self.samples = self.create_samples(
                self.rasters,
                self.rois,
                config,
                for_prediction)
            if self.out_dir is not None and not for_prediction:
                self.write_samples()
        elif config.app.algorithm in US_ALGORITHMS:
            SAMPLES_LOGGER.info("not implemented yet")
            sys.exit(1)
        else:
            self.samples = None

    def create_samples(self,
                       rasters: List[Path],
                       rois: List[dict],
                       config: Configuration,
                       for_prediction: bool) -> pd.DataFrame:
        """Create the entire dataset based on rasters and rois

        Args:
            rasters (List[Path]): list of raster file names
            rois (List[Dict]): contains each roi as a single dict
            config (Configuration): contains config
            for_prediction (bool): changes imputation/samples deleting behaviour

        Returns:
            samples (pd.DataFrame): dataset with samples
        """

        # Extract raster dates from file names
        raster_dates = [get_raster_date(raster) for raster in rasters]
        dates = [datetime.datetime.strptime(
            x, "%Y-%m-%d") for x in raster_dates]

        # Count bands and create new band names
        b_count = count_bands(rasters[0])
        bands = [f'B{str(x).zfill(2)}' for x in np.arange(1, b_count + 1)]

        samples = self.gather_samples_from_data(
            rasters, rois, bands, dates, config, for_prediction)
        return samples

    @ staticmethod
    def impute_ts(samples_df: pd.DataFrame) -> pd.DataFrame:
        """Imputation of timeseries
           Imputes values through time.

            Args:
                samples_df (pd.DataFrame): Dataframe with timeseries samples

            Returns:
                samples_filled_df (pd.DataFrame): Imputed DataFrame
        """

        # Shape after: rows:dates(time)   columns: class, roi_fid, pixel
        samples_df_time_as_index = samples_df.T.unstack()

        # Interpolate through time
        samples_df_filled = samples_df_time_as_index.interpolate(
            method="time", axis=0, limit_direction="both")

        # if there are still NaNs, there were not enough values to interpolate
        if samples_df_filled.isnull().values.any():
            samples_df_filled = samples_df_filled.fillna(
                method="ffill", axis=0).fillna(
                    method="bfill", axis=0)

        # Convert back to old shape
        samples_df_filled = samples_df_filled.T.unstack()

        return samples_df_filled

    @ staticmethod
    def check_geo_input_data_ts(rasters: List[Path]) -> None:
        """Check the rasters and rois input data

        Args:
            rasters (List[Path]): Raster paths
        """

        # check if the timeseries rasters have same amount of bands
        _ = verify_and_count_bands(rasters)



"""Vector utilities"""
import logging
from pathlib import Path
from typing import Any

import geopandas as gpd
import numpy as np
import fiona
import rasterio
from geopandas import GeoDataFrame
from rasterio import Affine
from rasterio.features import shapes
from rasterio.windows import Window
from shapely.geometry import MultiPolygon, Polygon

UTILS_VECTOR_LOGGER = logging.getLogger(__name__)


def vectorize_array(
        array_to_vectorize: np.ndarray,
        transform: Affine = None) -> GeoDataFrame:
    """Vectorize a tif file and return geodataframe.

    Args
        array_to_vectorize (array): Array to vectorize_and_classify
        transform (Affine): Affine transformation

    Returns:
        GeoDataFrame: geodataframe with polygons

    """
    polygonized = shapes(array_to_vectorize,
                         connectivity=8,
                         transform=transform)
    shapelist, valuelist = zip(*polygonized)
    shapelist = [Polygon(x['coordinates'][0]) for x in shapelist]
    new_poly_dict = {'geometry': shapelist, 'segment_id': valuelist}

    return gpd.GeoDataFrame.from_dict(new_poly_dict)


def get_subset_of_polygons(
        window: Window, transform: dict, polygons: GeoDataFrame,
        col_name: str = 'max') -> list:
    """Get a subset of a geodataframe from a rasterio window

    Uses a rasterio window to make an intersection on the whole polygon
    dataframe.

    Args:
        window (rasterio.Window): window of interest
        transform (dict):   transform of the window
        polygons (GeoDataFrame): polygons to classify
        col_name (str): Name of the column to subset

    Returns:
        list of [geometry, value] pairs that can be consumed by GeoPandas

    """
    left, bottom, right, top = rasterio.windows.bounds(window,
                                                       transform)

    window_polygon = Polygon([[left, bottom],
                              [left, top],
                              [right, top],
                              [right, bottom]])

    window_df = gpd.GeoDataFrame({'geometry': [window_polygon],
                                  'name': [1]})
    window_df.crs = {'init': 'epsg:4326'}

    subset_polygons = gpd.overlay(window_df,
                                  polygons,
                                  how='intersection')

    return list(zip(list(subset_polygons['geometry'].values),
                    list(subset_polygons[col_name].values)))


def gdf_polygon_to_multipolygon(gdf: GeoDataFrame) -> GeoDataFrame:
    """Converts all geometries to multipolyon, so there are no issues when
    saving these to a file and adds a crs (epsg:4326) to the gdf

    Args:
        gdf (GeoDataFrame): Containing polygons

    Returns:
        gdf (GeoDataFrame): GDF with all polygons converted to multipolygons

    """
    geometry = [MultiPolygon([feature]) if isinstance(feature, Polygon)
                else feature for feature in gdf["geometry"]]
    return geometry


def get_rois_extent(rois: Path) -> Any:
    """

    Args:
        rois (Path): path to a rois file

    Returns:
        bounds of rois (list): ulx, uly, llx, lly coordinates

    """
    with fiona.open(rois, "r") as shapefile:
        return shapefile.bounds



""" Unittests for config file"""
from classifier.utils.config import (
    AccuracyConfiguration, AppConfiguration, Configuration,
    OptimizeSupervisedConfiguration, RandomForestConfiguration,
    SupervisedConfiguration, UnsupervisedConfiguration, json_to_config)
from classifier.utils.config import DtwConfiguration, json_to_config
import copy
import json
import os
import unittest

from classifier.utils.config import PARAMETERS
from classifier.settings import WORKSPACE


class ConfigTestCase(unittest.TestCase):
    """ Testclass for config.py"""

    @classmethod
    def setUp(cls):
        algorithm = "randomforest"
        window = 1024
        model = None
        model_save = False  # Save a model file which can be re-ed
        samples = None
        log_level = "INFO"  # Logging level
        threads = -1  # of threads
        imputation = True  # e simple imputation for missing values
        imputation_strategy = "mean"  # Strategy for imputation. mean,
        rasters_are_timeseries = False  # Rasters are timeseries
        imputation_constant = -9999  # constant for imputation
        # if constant was chosen as imputation method
        probability = True
        all_probabilities = False
        optimize = False  # Optimize the model parameters
        optimize_number = 10  # Number of iterations for optimization
        remove_outliers = True  # Remove outliers from the training data
        nclasses = 2  # Number of classes for unsupervised
        trainfraction = 1.0  # Fraction of raster ed for training
        perform_assesment = True  # Perform accuracy assessment
        testfraction = 0.25  # Fraction of data to e for training

        cls.ref_config = Configuration(
            AppConfiguration(algorithm=algorithm,
                             window=window,
                             model=model,
                             model_save=model_save,
                             samples=samples,
                             log_level=log_level,
                             threads=threads,
                             imputation=imputation,
                             imputation_strategy=imputation_strategy,
                             rasters_are_timeseries=rasters_are_timeseries,
                             imputation_constant=imputation_constant),
            SupervisedConfiguration(
                probability=probability,
                all_probabilities=all_probabilities,
                remove_outliers=remove_outliers,
                optimization=OptimizeSupervisedConfiguration(
                    optimize=optimize,
                    optimize_number=optimize_number),
            ),
            UnsupervisedConfiguration(nclasses, trainfraction),
            AccuracyConfiguration(perform_assesment, testfraction),
            DtwConfiguration(
                patterns=None,
                patterns_save=False,
                number_of_patterns_per_class=1,
                n_neighbors=1,
                use_gam=True,
                window=None,
                max_dist=None,
                use_pruning=False,
                penalty=None,
                optimization_parameters={
                    "n_neighbors": [1],
                    "number_of_patterns_per_class": [1],
                    "use_gam": [True],
                    "window": [None],
                    "penalty": [None]
                }
            ),
            RandomForestConfiguration(
                n_estimators=100,
                criterion="gini",
                max_depth=None,
                min_samples_split=2,
                min_samples_leaf=1,
                min_weight_fraction_leaf=0.0,
                max_features="auto",
                max_leaf_nodes=None,
                min_impurity_decrease=0.0,
                bootstrap=True,
                oob_score=False,
                random_state=None,
                verbose=0,
                class_weight=None,
                ccp_alpha=0.0,
                max_samples=None,
                optimization_parameters={
                    "max_features": ['auto', 'sqrt', 'log2'],
                    "max_leaf_nodes": [3, 5, 7],
                    "max_depth": [None, 1, 3, 10, 20000],
                    "n_estimators": [10, 50, 100],
                    "min_samples_split": [2],
                    "min_samples_leaf": [1],
                    "min_weight_fraction_leaf": [0.0],
                    "min_impurity_decrease": [0.0],
                    "oob_score": [False],
                    "class_weight": [None],
                    "ccp_alpha": [0.0],
                    "max_samples": [None]
                }
            ),
            name='',
            tmp_dir=None
        )

    def test_read_config_file_outputasexpected(self):
        """Config file is correctly read in"""
        params = copy.deepcopy(PARAMETERS)
        config = json_to_config(json.dumps(params))
        self.assertEqual(self.ref_config, config)

    def test_read_config_file__raises(self):
        """Raises error due to invalid config file"""
        params = copy.deepcopy(PARAMETERS)
        params['app']['model'] = "C:///:/Documen"

        with self.assertRaises(SystemExit):
            json_to_config(json.dumps(params))

    def test_read_config_file__model_not_zip_raises(self):
        """Raises error due to invalid config file
        model not .zip"""
        params = copy.deepcopy(PARAMETERS)
        params['app']['model'] = os.path.join(WORKSPACE, "model")

        with self.assertRaises(SystemExit):
            json_to_config(json.dumps(params))

    def test_read_config_file__samples_not_csv_or_pkl_raises(self):
        """Raises error due to invalid config file
        samples is not .pkl or .csv"""
        params = copy.deepcopy(PARAMETERS)
        params['app']['samples'] = os.path.join(WORKSPACE, "samples.jpg")

        with self.assertRaises(SystemExit):
            json_to_config(json.dumps(params))
